use serde::{Deserialize, Serialize};
use std::{
    collections::{hash_map::Entry, HashMap},
    fmt::Display,
    sync::{Arc, Mutex},
};
use tokio::sync::mpsc;

use egui::{
    load::{ImageLoader, ImagePoll, LoadError},
    ColorImage, Context, Vec2,
};

use crate::emulator::{EmulatorClient, EmulatorCommand, SimId};

enum VramRequest {
    Load(String, VramResource),
}

enum VramResponse {
    Loaded(String, ColorImage),
}

pub struct VramResource {
    sim: SimId,
    kind: VramResourceKind,
}

impl VramResource {
    pub fn character_data(sim: SimId, palette: VramPalette) -> Self {
        Self {
            sim,
            kind: VramResourceKind::CharacterData { palette },
        }
    }

    pub fn character(sim: SimId, palette: VramPalette, index: usize) -> Self {
        Self {
            sim,
            kind: VramResourceKind::Character { palette, index },
        }
    }

    pub fn palette_color(sim: SimId, palette: VramPalette, index: usize) -> Self {
        Self {
            sim,
            kind: VramResourceKind::PaletteColor { palette, index },
        }
    }

    pub fn to_uri(&self) -> String {
        format!(
            "vram://{}:{}",
            self.sim.to_index(),
            serde_json::to_string(&self.kind).unwrap(),
        )
    }

    fn from_uri(uri: &str) -> Option<VramResource> {
        let uri = uri.strip_prefix("vram://")?;
        let (sim, uri) = match uri.split_at_checked(2)? {
            ("0:", rest) => (SimId::Player1, rest),
            ("1:", rest) => (SimId::Player2, rest),
            _ => return None,
        };
        let kind = serde_json::from_str(uri).ok()?;
        Some(Self { sim, kind })
    }
}

#[derive(Serialize, Deserialize)]
enum VramResourceKind {
    Character { palette: VramPalette, index: usize },
    CharacterData { palette: VramPalette },
    PaletteColor { palette: VramPalette, index: usize },
}

impl VramResourceKind {
    fn size(&self) -> Option<Vec2> {
        match self {
            Self::Character { .. } => Some((8.0, 8.0).into()),
            Self::CharacterData { .. } => Some((8.0 * 16.0, 8.0 * 128.0).into()),
            Self::PaletteColor { .. } => Some((1.0, 1.0).into()),
        }
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Serialize, Deserialize)]
pub enum VramPalette {
    Generic,
    Bg0,
    Bg1,
    Bg2,
    Bg3,
    Obj0,
    Obj1,
    Obj2,
    Obj3,
}

impl VramPalette {
    pub const fn values() -> [VramPalette; 9] {
        [
            Self::Generic,
            Self::Bg0,
            Self::Bg1,
            Self::Bg2,
            Self::Bg3,
            Self::Obj0,
            Self::Obj1,
            Self::Obj2,
            Self::Obj3,
        ]
    }
}

impl Display for VramPalette {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::Generic => f.write_str("Generic"),
            Self::Bg0 => f.write_str("BG 0"),
            Self::Bg1 => f.write_str("BG 1"),
            Self::Bg2 => f.write_str("BG 2"),
            Self::Bg3 => f.write_str("BG 3"),
            Self::Obj0 => f.write_str("OBJ 0"),
            Self::Obj1 => f.write_str("OBJ 1"),
            Self::Obj2 => f.write_str("OBJ 2"),
            Self::Obj3 => f.write_str("OBJ 3"),
        }
    }
}

pub struct VramLoader {
    cache: Mutex<HashMap<String, ImagePoll>>,
    source: Mutex<mpsc::UnboundedReceiver<VramResponse>>,
    sink: mpsc::UnboundedSender<VramRequest>,
}

impl VramLoader {
    pub fn new(client: EmulatorClient) -> Self {
        let (tx1, rx1) = mpsc::unbounded_channel();
        let (tx2, rx2) = mpsc::unbounded_channel();
        std::thread::spawn(move || {
            tokio::runtime::Builder::new_current_thread()
                .enable_all()
                .build()
                .unwrap()
                .block_on(async move {
                    let worker = VramLoadingWorker::new(rx2, tx1, client);
                    worker.run().await;
                })
        });
        Self {
            cache: Mutex::new(HashMap::new()),
            source: Mutex::new(rx1),
            sink: tx2,
        }
    }
}

impl ImageLoader for VramLoader {
    fn id(&self) -> &str {
        concat!(module_path!(), "::VramLoader")
    }

    fn load(
        &self,
        _ctx: &Context,
        uri: &str,
        _size_hint: egui::SizeHint,
    ) -> Result<ImagePoll, LoadError> {
        let Some(resource) = VramResource::from_uri(uri) else {
            return Err(LoadError::NotSupported);
        };
        let mut cache = self.cache.lock().unwrap();
        {
            let mut source = self.source.lock().unwrap();
            while let Ok(response) = source.try_recv() {
                match response {
                    VramResponse::Loaded(uri, image) => {
                        cache.insert(
                            uri,
                            ImagePoll::Ready {
                                image: Arc::new(image),
                            },
                        );
                    }
                }
            }
        }
        let poll = match cache.entry(uri.to_string()) {
            Entry::Occupied(entry) => entry.into_mut(),
            Entry::Vacant(entry) => {
                let pending = ImagePoll::Pending {
                    size: resource.kind.size(),
                };
                let _ = self.sink.send(VramRequest::Load(uri.to_string(), resource));
                entry.insert(pending)
            }
        };
        Ok(poll.clone())
    }

    fn forget(&self, uri: &str) {
        self.cache.lock().unwrap().remove(uri);
    }

    fn forget_all(&self) {
        self.cache.lock().unwrap().clear();
    }

    fn byte_size(&self) -> usize {
        self.cache
            .lock()
            .unwrap()
            .values()
            .map(|c| match c {
                ImagePoll::Pending { .. } => 0,
                ImagePoll::Ready { image } => image.pixels.len() * size_of::<ColorImage>(),
            })
            .sum()
    }
}

struct VramLoadingWorker {
    source: mpsc::UnboundedReceiver<VramRequest>,
    sink: mpsc::UnboundedSender<VramResponse>,
    client: EmulatorClient,
}

impl VramLoadingWorker {
    fn new(
        source: mpsc::UnboundedReceiver<VramRequest>,
        sink: mpsc::UnboundedSender<VramResponse>,
        client: EmulatorClient,
    ) -> Self {
        Self {
            source,
            sink,
            client,
        }
    }
    async fn run(mut self) {
        while let Some(request) = self.source.recv().await {
            #[allow(irrefutable_let_patterns)]
            if let VramRequest::Load(uri, resource) = request {
                let Some(image) = self.load(resource).await else {
                    continue;
                };
                if self.sink.send(VramResponse::Loaded(uri, image)).is_err() {
                    return;
                }
            }
        }
    }

    async fn load(&self, resource: VramResource) -> Option<ColorImage> {
        let sim = resource.sim;
        match resource.kind {
            VramResourceKind::Character { palette, index } => {
                self.load_character(sim, palette, index).await
            }
            VramResourceKind::CharacterData { palette } => {
                self.load_character_data(sim, palette).await
            }
            VramResourceKind::PaletteColor { palette, index } => {
                self.load_palette_color(sim, palette, index).await
            }
        }
    }

    async fn load_character(
        &self,
        sim: SimId,
        palette: VramPalette,
        index: usize,
    ) -> Option<ColorImage> {
        if index >= 2048 {
            return None;
        }
        let address = 0x00078000 + (index as u32 * 16);
        let (memory, palette) = tokio::join!(
            self.read_memory(sim, address, 16),
            self.load_palette_colors(sim, palette),
        );
        let palette = palette?;
        let mut buffer = vec![];
        for byte in memory? {
            for offset in (0..8).step_by(2) {
                let char = (byte >> offset) & 0x3;
                buffer.push(palette[char as usize]);
            }
        }
        Some(ColorImage::from_gray([8, 8], &buffer))
    }

    async fn load_character_data(&self, sim: SimId, palette: VramPalette) -> Option<ColorImage> {
        let (memory, palette) = tokio::join!(
            self.read_memory(sim, 0x00078000, 16 * 2048),
            self.load_palette_colors(sim, palette),
        );
        let palette = palette?;
        let mut buffer = vec![0; 8 * 8 * 2048];
        for (i, byte) in memory?.into_iter().enumerate() {
            let bytes = [0, 2, 4, 6].map(|off| palette[(byte as usize >> off) & 0x3]);
            let char_index = i / 16;
            let in_char_pos = i % 16;
            let x = ((char_index % 16) * 8) + ((in_char_pos % 2) * 4);
            let y = ((char_index / 16) * 8) + (in_char_pos / 2);
            let write_index = (y * 16 * 8) + x;
            buffer[write_index..(write_index + 4)].copy_from_slice(&bytes);
        }
        Some(ColorImage::from_gray([8 * 16, 8 * 128], &buffer))
    }

    async fn load_palette_color(
        &self,
        sim: SimId,
        palette: VramPalette,
        index: usize,
    ) -> Option<ColorImage> {
        if index == 0 {
            return Some(ColorImage::from_gray([1, 1], &[0]));
        }
        if index > 3 {
            return None;
        }
        let shade = *self.load_palette_colors(sim, palette).await?.get(index)?;
        Some(ColorImage::from_gray([1, 1], &[shade]))
    }

    async fn load_palette_colors(&self, sim: SimId, palette: VramPalette) -> Option<[u8; 4]> {
        let offset = match palette {
            VramPalette::Generic => {
                return Some([0, 64, 128, 255]);
            }
            VramPalette::Bg0 => 0,
            VramPalette::Bg1 => 2,
            VramPalette::Bg2 => 4,
            VramPalette::Bg3 => 6,
            VramPalette::Obj0 => 8,
            VramPalette::Obj1 => 10,
            VramPalette::Obj2 => 12,
            VramPalette::Obj3 => 14,
        };
        let (palettes, brightnesses) = tokio::join!(
            self.read_memory(sim, 0x0005f860, 16),
            self.read_memory(sim, 0x0005f824, 6),
        );
        let palette = *palettes?.get(offset)?;
        let brts = brightnesses?;
        let shades = [
            0,
            brts[0],
            brts[2],
            brts[0].saturating_add(brts[2]).saturating_add(brts[4]),
        ];
        Some([
            0,
            shades[(palette >> 2) as usize & 0x03],
            shades[(palette >> 4) as usize & 0x03],
            shades[(palette >> 6) as usize & 0x03],
        ])
    }

    async fn read_memory(&self, sim: SimId, address: u32, size: usize) -> Option<Vec<u8>> {
        let (tx, rx) = oneshot::channel();
        self.client
            .send_command(EmulatorCommand::ReadMemory(sim, address, size, vec![], tx));
        rx.await.ok()
    }
}
